/*******************************************************************************
Copyright (c) 2004 Oracle Corporation/AMIS Services BV

File        : GetDataObject.java

Notes       :

Open Issues :

$revision_history$
19-may-2004   Lucas Jellema
  1.16AMIS1.0  - Created this action based on the 1.16 version of GetDataObject that Oracle will ship in JHeadstart 10g 9.0.5.1 (late May/early June 2004)
              - small changes: mapping.getRefresh is not yet available; this class will never requery a DataObject that is already in the session data
03-mei-2004   Peter Ebell
  1.16        - If noDataObjectFound, still create the bean when necessary and clear its contents,
                to prevent an old DataObject to appear (or a nullpointer if the bean was not present yet.
              - If keyParameter is not found on request, do not give nullpointer but treat it in the same
                manner as if it was present but had a null value.
16-feb-2004   Peter Ebell
  1.15        Added refresh feature, so that we can prevent that a DataObject will be retrieved
              if it already exists in the SessionData.
22-okt-2003   Peter Ebell
  1.14        For keyParameterName, do not assume that the resulting object is a String (as it may
              be nested, it might be that actual 'domain' values are retrieved. So, we now first
              check if it is a String, and if not, we call the toString method.
09-sep-2003   Steven Davelaar
  1.13        
09-sep-2003   Steven Davelaar
  1.12        Use getNestedAttribute instead to retrieve keyParameter
07-sep-2003   Steven Davelaar
  1.10        Improved javadoc on queryType
02-sep-2003   Steven Davelaar
  1.9         Added boolean argument convertValue to call to getNestedAttribute
01-sep-2003   Peter Ebell
  1.8         Added javadoc for attributeBeanName property
01-sep-2003   Peter Ebell
  1.7         Call version of getNestedAttribute on SessionData that uses attributeBeanName
              parameter when constructing the map with query arguments,so that it only looks
              in the specified bean for the attribute, and not in the request. If attributeBeanName
              parameter was not supplied, SessionData will fall back to the version that does
              look in the request.
23-may-2003   Peter Ebell
  1.6         Use getDataObject instead of getObject in retrieveComplex
05-feb-2003   Steven Davelaar
  1.5         Use resultScope or scope property to store action result
              (if not present, store on session)
19-dec-2002   Steven Davelaar
  1.4         Fixed bug: check for empty dataObject before setting updatability
09-dec-2002   Steven Davelaar
  1.3         Added method validateProperties
13-nov-2002   Steven Davelaar
  1.2         Added support for readOnlyBean mapping property
11-nov-2002   Steven Davelaar
  1.1         Fixed bug, attribute did not check request object
22-oct-2002   Steven Davelaar
  1.0         Created based on oracle.jheadstart.process.GetDataObject
  
******************************************************************************/
/*
$add_to_revision_history
$
*/
package nl.amis.jheadstart.controller.struts.action;

// Java imports
import java.util.List;
import java.util.HashMap;

//struts imports
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;

// JHeadstart imports
import oracle.jheadstart.model.DataObjectHandler;

import oracle.jheadstart.exception.UserException;
import oracle.jheadstart.model.DataObject;
import oracle.jheadstart.JhsConstants;
import oracle.jheadstart.controller.struts.action.JhsAction;
import oracle.jheadstart.controller.struts.config.JhsActionMapping;
import oracle.jheadstart.controller.SessionData;
import oracle.jheadstart.view.struts.JhsDynaActionForm;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
  * <h3>Description</h3>
  * This action retrieves a single DataObject from the handler that belongs
  * to the dataObject interface that is specified as mapping property.
  * The DataObject is retrieved through the primary key when the <code>
  * keyParameterName</code> mapping property is specified, or through a custom
  * query when the <code>queryType</code> and, optionally, <code>attributeNames</code> 
  * mapping property is specified.
  * <br/>
  * <br/>
  * This action differs from the standard GetDataObject action in that it
  * does only try to get the DataObject from the Model if it does not already 
  * exist in the SessionData. If it is already there, it is not retrieved again
  * from the Model.
  * <br/>
  * 
  * <h3>Action Mapping Properties</h3>
  * <table BORDER="1" CELLPADDING="3" CELLSPACING="0" COLS=6 WIDTH="100%" >
  * <tr BGCOLOR="#CCCCFF" CLASS="TableHeadingColor">
  *  <td WIDTH="1%" VALIGN="top"><b>Name</b></td>
  *  <td WIDTH="1%" VALIGN="top"><b>Type</b></td>
  *  <td WIDTH="1%" VALIGN="top"><b>Required</b></td>
  *  <td WIDTH="1%" VALIGN="top"><b>Default Value</b></td>
  *  <td WIDTH="1%" VALIGN="top"><b>Allowable Values</b></td>
  *  <td WIDTH="45%" VALIGN="top"><b>Description</b></td>
  * </tr>
  *
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">dataObjectName</td><td VALIGN="top">String</td>
  * <td VALIGN="top">yes</td><td VALIGN="top">&nbsp;</td><td VALIGN="top">
  * &nbsp;
  * </td><td VALIGN="top">
  * Name that must be used to store the dataObject retrieved 
  * in the SessionData object.
  * </td></tr>
  *
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">dataObjectInterface</td><td VALIGN="top">String</td>
  * <td VALIGN="top">yes</td><td VALIGN="top">&nbsp;</td><td VALIGN="top">
  * &nbsp;
  * </td><td VALIGN="top">
  * Name of dataObject interface used to retrieve the handler that is used
  * to get the dataObject
  * </td></tr>
  *
  *
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">keyParameterName</td><td VALIGN="top">String</td>
  * <td VALIGN="top">no</td><td VALIGN="top">&nbsp;</td><td VALIGN="top">
  * &nbsp;
  * </td><td VALIGN="top">
  * The name of the SessionData attribute holding the value of
  * the primary key column for the dataObject to be retrieved. This value is passed
  * as argument to the getObject method called on the handler to retrieve the
  * dataObject. You can refer to a nested attributes by using a "." in the 
  * attribute name. 
  * </td></tr>
  *
  *
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">readOnlyBean</td><td VALIGN="top">String</td>
  * <td VALIGN="top">no</td><td VALIGN="top">&nbsp;</td><td VALIGN="top">
  * &nbsp;
  * </td><td VALIGN="top">
  * If this property is specified, a <code>DynaBean</code> of class 
  * <code>oracle.jheadstart.ui.data.DynaBeanReadOnly</code>
  * is created and stored in the SessionData. The browsed DataObject is
  * set on the bean. The bean can be used to display data in a JSP using
  * the Jakarta Struts html and bean tag libraries. By using this property,
  * you do not need to define a &lt;form-bean&gt; element in the struts-config
  * file. However, the bean is read-only, the values entered in
  * the JSP page are not set on the bean.
  * </td></tr>
  * 
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">isUpdatable</td><td VALIGN="top">Boolean</td>
  * <td VALIGN="top">no</td><td VALIGN="top">false</td><td VALIGN="top">
  * true, false
  * </td><td VALIGN="top">
  * This property setting determines the values returned by the transient 
  * dataObject attributes isUpdating, isNotUpdating, isViewing and isNotViewing.
  * When using the same UIX or JSP page to insert, update or view a dataObject,
  * these transient attribute values (together with isInserting and isNotInserting)
  * can be used to determine whether a persistent dataObject attribute
  * should be displayed as an enterable Text or GUI item or as read-only text. 
  * </td></tr>
  *
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">queryType</td><td VALIGN="top">String</td>
  * <td VALIGN="top">no</td><td VALIGN="top">&nbsp;</td><td VALIGN="top">
  * &nbsp;
  * </td><td VALIGN="top">
  * Identifies the type of query to be used. The queryType defines which
  * ViewObject (BC4J) or named query (TopLink) is used to query the database.
  * The ViewObject or named query can have bind parameters. In the 
  * <code>attributeNames</code>
  * property, you can specify which SessionData atributes (or request
  * parameters), must be used to retrieve the bind parameter values.
  * Note that the number and sequence of the query arguments must match the 
  * number and sequence of the bind parameters defined in the query. You also
  * must set the property <code>useAttributeNameAsKey</code> to false for this 
  * automatic parameter binding to work.
  * </td></tr>
  *
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">attributeNames</td><td VALIGN="top">String</td>
  * <td VALIGN="top">no</td><td VALIGN="top">&nbsp;</td><td VALIGN="top">
  * &nbsp;
  * </td><td VALIGN="top">
  * Comma separated list with SessionData attribute names (or request
  * parameters) that should be used as query parameter. You can refer
  * to nested attributes by using a "." in the attribute name. 
  * </td></tr>
  *
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">useAttributeNameAsKey</td><td VALIGN="top">String</td>
  * <td VALIGN="top">no</td><td VALIGN="top">true</td><td VALIGN="top">
  * true, false
  * </td><td VALIGN="top">
  * Indicates whether attribute values are keyed by name or index number in the
  * HashMap that is passed to the getDataObject() method on the handler. Index number
  * starts with 0, and is incremented with 1 for each attribute included in the
  * attributeNames property. Set this property to false, if you want to
  * leverage the default BC4J or TopLink handler implementation of getDataObjectSet() method
  * with query type. If you set this property to true, you must provide your own
  * handler implementation of the query. To do this, you should override method
  * <code>getDataObject(queryType,arguments,sessionId)</code>,
  * check for the queryType passed in, and construct your query, using the 
  * attributeNames and their values that are passed in as a hashmap.
  * </td></tr>
  *
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">attributeBeanName</td><td VALIGN="top">String</td>
  * <td VALIGN="top">no</td><td VALIGN="top">&nbsp;
  * </td><td VALIGN="top">&nbsp;
  * </td><td VALIGN="top">
  * If specified, indicates that the attributes
  * listed in the <code>attributeNames</code> mapping property are not to be
  * taken from the request, but from a Form bean with this name.
  * </td></tr>
  *
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">refresh</td><td VALIGN="top">Boolean</td>
  * <td VALIGN="top">no</td><td VALIGN="top">true
  * </td><td VALIGN="top">true, false
  * </td><td VALIGN="top">
  * When set to true (the default), the DataObject will always be retrieved
  * from the Model and stored in the SessionData, even if an existing DataObject
  * already existed with the specified name. When set to false, this will only occur
  * if no DataObject with the specified name already existed in the SessionData.
  * </td></tr>
  * 
  * <tr BGCOLOR="white" CLASS="TableRowColor">
  * <td VALIGN="top">resultScope</td><td VALIGN="top">String</td>
  * <td VALIGN="top">no</td><td VALIGN="top">defaults to value of scope property, which in turn defaults to 'session'</td><td VALIGN="top">
  * request, session, application
  * </td><td VALIGN="top">
  * This property sets the scope used to store the DataObject in
  * the SessionData.
  * </td></tr>
  *
  * </table>
  * <h3>Action Parameters</h3>
  * At least one of the following parameters must be present in SessionData
  *   <ul>
  *     <li><code>[keyParameterName]</code>: The name of this action parameter is 
  *     determined dynamically at runtime based on the value passed in to the
  *     <code>keyParameterName</code> mapping property.  It is the name of
  *     the object in the SessionData containing the value of the primary key
  *     to query on.</li>
  *     <li><code>[attributeNames]</code>: If the <code>keyParameterName</code> 
  *     property is not specified, then the <code>attributeNames</code> property
  *     must contain a comma-delimited list of names of SessionData objects
  *     that contain the values to be used in the query.
  *   </ul>
  *
  * <h3>Action Results</h3>
  * The following results are stored in the SessionData by this action
  *   <ul>
  *     <li><code>[dataObjectName]</code>: This action result will
  *     contain the single DataObject retrieved by this action.
  *     <p>
  *     If the action is associated with a dynamic action form (set through the 
  *     standard Struts <code>name</code> mapping property), and the
  *     form bean is of type <code>JhsDynaActionForm</code>, the DataObject is
  *     set on the form bean, using method <code>setDataObject()</code>.
  *     </li>
  *   </ul>
  * The scope used to store the results is taken from the <code>resultScope</code>
  * mapping property. If this property is not set, the scope is taken
  * from the (standard Struts) <code>scope</code> mapping property. If this 
  * property is not set, the scope defaults to session. 
  *
  * <h3>User Exceptions</h3>
  * The following User Exceptions might be thrown by this action
  *   <ul>
  *     <li><code>JHS-00102</code>: Please select a [dataObjectName].  This 
  *     exception is thrown when the <code>keyParameterName</code> is not
  *     found in the SessionData.</li>
  *   </ul>
  *
  * <h3>Action Forwards</h3>
  * When using this action, you should include forwards in the Struts-config
  * file for all forwards that can be returned by this action:
  *   <ul>
  *     <li><code>success</code>: set when a DataObject is found and stored 
  *     in the SessionData</li>
  *     <li><code>failure</code>: set when unexpected exception is encountered</li>
  *     <li><code>userError</code>: set when the neither a ResourceId or a set
  *     of arguments was specified</li>
  *     <li><code>dataObjectNotFound</code>: set when no records matching the requested
  *     primary key or search criteria were found</li>
  *   </ul>
  *
  * 
  * <h3>Communication with Model Layer</h3>
  * The following methods can be called on interface 
  * <code>oracle.jheadstart.model.DataObjectHandler</code>:
  * <ul>
  *  <li><a href="../../../model/DataObjectHandler.html#getDataObject(java.lang.Object, java.lang.Object)">
  *      <code>getDataObject(Object id, Object sessionId)</code></a> called when 
  *      mapping property <code>keyParameterName</code> is not null. 
  *  </li>
  *  <li><a href="../../../model/DataObjectHandler.html#getDataObject(java.lang.String, java.util.HashMap, java.lang.Object)">
  *      <code>getDataObject(String queryType,
  *                          HashMap arguments,
  *                          Object sessionId)</code></a> called when 
  *      mapping property <code>keyParameterName</code> is null.
  *  </li>
  * </ul>
  * The following method is called on interface 
  * <code>oracle.jheadstart.model.DataObject</code>:
  * <ul>
  *  <li><a href="../../../model/DataObject.html#setIsUpdatable(boolean)">
  *      <code>setIsUpdatable(boolean isUpdatable)</code></a> argument value taken
  *      from mapping property isUpdatable
  *  </li>
  * </ul>
  *
  * @author Steven Davelaar
  */

public class GetDataObject extends JhsAction
{

  protected Log mLog = LogFactory.getLog(GetDataObject.class);     


  /**
   * Start execution of action logic.
   * <p>
   * By using this method instead of the standard Struts execute
   * method we get easy access to JHeadstart Action Mappings
   * and JHeadstart SessionData as well as default handling of exceptions.
   * In addition, by coding against the SessionData interface instead of directly
   * against HttpServletRequest and HttpServletResponse, the action
   * become less Servlet specific and can easier be reused in combination
   * with future non-(http)servlet controllers.
   * @see JhsAction#execute(org.apache.struts.action.ActionMapping mapping, org.apache.struts.action.ActionForm form, javax.servlet.http.HttpServletRequest request, javax.servlet.http.HttpServletResponse response)
   */
  public ActionForward execute(JhsActionMapping mapping
                         ,SessionData sessionData 
                         ,ActionForm form) 
    throws UserException, Exception

  {
    DataObject dataObject = null;
    
    // Check if the dataObject is already present. If so,
    // just return the success forward.
    // NOTE: could do some more advanced checking (same dataObjectInterface etc).
      dataObject = sessionData.getDataObject(mapping.getDataObjectName());
      if (dataObject != null)
      { 
        mLog.debug("Not retrieving DataObject \""+mapping.getDataObjectName()+"\" because refresh=false, and DataObject is already present");
        return mapping.findForward(JhsConstants.SUCCESS);
      }
    
    if (mapping.getKeyParameterName() == null)
    {
      dataObject = retrieveComplex(mapping, sessionData);
    }
    else
    {
      dataObject = retrieveSimple(mapping, sessionData);
    }

    // Even if the dataObject is null, we want to create and clear the form beans
    // to prevent the presence of an older version of the DataObject to be available
    // through the form beans
    if (mapping.getReadOnlyBean()!=null) 
    {
      createReadOnlyBean(sessionData,mapping.getReadOnlyBean()
                        ,dataObject, getResultScope(mapping));
    }
    else if (form!=null && form instanceof JhsDynaActionForm)
    {
      ((JhsDynaActionForm)form).setDataObject(dataObject);
    }

    if (dataObject == null)
    {
      // remove a previously existing DataObject if it's there
      sessionData.removeAttribute(mapping.getDataObjectName(), getResultScope(mapping));
      return mapping.findForward(JhsConstants.DATA_OBJECT_NOT_FOUND);
    }


    // set updatable property on dataObject. This is used by the dataObject to
    // return the correct values for the dml action transient attributes
    // (isInserting, isUpdating, isViewing, etc.)
    // Value of updatable is determined as follows: the Updatable specified
    // through mapping property takes precedence, otherwise it is set to
    // 'true'.
    if (mapping.getUpdatableSpecified().booleanValue())
    {
      dataObject.setIsUpdatable(mapping.getIsUpdatable().booleanValue());
    }
    else
    {
      dataObject.setIsUpdatable(true);
    }
    sessionData.setAttribute(mapping.getDataObjectName()
                          , dataObject
                          , getResultScope(mapping));

    return mapping.findForward(JhsConstants.SUCCESS);
  }

  /* Retrieve the dataObject with a simple query based on the ID
    */
  protected DataObject retrieveSimple(JhsActionMapping mapping
                                     , SessionData sessionData) 
    throws Exception
  {
    // retrieve Session Id from InfoTable
    Object sessionId = sessionData.getSessionId();

    // Get the Resource handler
    DataObjectHandler handler =
      (DataObjectHandler)getHandler(mapping.getDataObjectInterface(), sessionData);


    // retrieve Resource Id from session data
    String idstring = null;

    Object idObject = sessionData.getNestedAttribute(mapping.getKeyParameterName(),true);
    if (idObject instanceof String)
      idstring = (String)idObject;
    else if (idObject != null)
      idstring = idObject.toString();
      
    if (idstring==null)
    {
      String dataObjectName = handler.getDisplayName();
      if (dataObjectName==null)
      {
        dataObjectName = mapping.getDataObjectName();
      }
      Object[] params = {dataObjectName};
      throw new UserException("JHS-00102",params,"Please select a {0}",ERROR );
    }


    return (DataObject)handler.getDataObject(idstring, sessionId);
  }

  /* Retrieve the dataObject with a complex query.
    * A complex query is every query that is not based only on the Id column
    * This methods gets the values of all arguments from the InfoTable and puts
    * them in a List.
    * A List is used, rather than a Map, to get access if the argument
    * values by index. The handlerimplementation should not know about
    * the argumentnames used in a particular case.
    * Using a list makes it possible to have different argumentnames
    * in the servicexml file. The order of the arguments in the service xml file
    * should be the same as the handlerimplementation expects them.
    */
  protected DataObject retrieveComplex(JhsActionMapping mapping
                                     , SessionData sessionData) 
    throws Exception
  {

    List argumentNames = mapping.getAttributeNamesList();
    HashMap arguments = null;
    if (argumentNames!=null)
    {
      arguments = new HashMap(argumentNames.size(),1);
      String name =null;
    
      for (int i=0; i<argumentNames.size();i++)
      {
        // Get the value from the SessionData
        name=(String)argumentNames.get(i);
        Object value = sessionData.getNestedAttribute(mapping.getAttributeBeanName(),name, true);
        if (mapping.getUseAttributeNameAsKey().booleanValue())
        {
           arguments.put(name,value);
        }
        else
        {
           arguments.put(new Integer(i),value);        
        }
      }
    }
    // retrieve Session Id from InfoTable
    Object sessionId = sessionData.getSessionId();

    // Get the Resource handler
    DataObjectHandler handler =
      (DataObjectHandler)getHandler(mapping.getDataObjectInterface(), sessionData);

    return handler.getDataObject(mapping.getQueryType(),arguments,sessionId);
  }

  /**
   * Checks whether combination of properties and property values
   * specified for the action mappings of this action in the 
   * struts-config is valid.
   */
   protected void validateProperties(JhsActionMapping mapping
                                         ,SessionData sessionData 
                                         ,ActionForm form)
   {
     mapping.checkForwards(new String[]{SUCCESS, USER_ERROR, DATA_OBJECT_NOT_FOUND});
     // check required properties   
     mapping.checkRequiredProperty("dataObjectInterface",mapping.getDataObjectInterface());
     mapping.checkRequiredProperty("dataObjectName",mapping.getDataObjectName());

     // readOnlyBean and name cannot be set at same time, and readOnlyBean should
     // have name that differs from dataObjectName
     if (mapping.getReadOnlyBean()!=null)
     {
       if (mapping.getReadOnlyBean().equals(mapping.getDataObjectName()))
       {
         mapping.addInvalidPropertyException
           ("Properties dataObjectName and readOnlyBean cannot have "
            +" the same value: " + mapping.getReadOnlyBean());
       }
     }   }

}
